import urllib
import http.client
import json
import re
import ssl
import copyreg
# from multiprocessing.reduction import rebuild_socket, reduce_socket


class Request:

    def __init__(self):
        # copyreg.pickle(ssl.SSLSocket, reduce_socket, rebuild_socket)
        copyreg.pickle(ssl.SSLContext, self._save_sslcontext)
        pass

    def _save_sslcontext(obj):
        return obj.__class__, (obj.protocol,)

    def _parse_content_type(self, c_type):
        res = re.match(r"(?P<type>\w+?)\/(?P<app>\w+?);"
                       r"\s*charset=(?P<charset>.+)",
                       c_type)
        return res.groupdict()

    def _parse_body(self, raw, content_type):
        if content_type["app"] == "json":
            body = json.loads(raw)
        else:
            body = raw.decode(content_type["charset"])
        return body

    def _send(self, type, host, path, headers):
        try:
            self.https.host = host
        except AttributeError:
            self.https = http.client.HTTPSConnection(host)
        try:
            self.https.request(type, path, headers=headers)
        except http.client.CannotSendRequest:
            print("Cannot send request, refreshing connection")
            self.https = http.client.HTTPSConnection(host)
            self.https.request(type, path, headers=headers)
        # parse the response
        response = self.https.getresponse()
        # print("RESPONSE STATUS", response.status)
        content_type = self._parse_content_type(
            response.getheader("content-type")
        )
        self.response = response
        self.headers = response.getheaders()
        self.body = self._parse_body(response.read(), content_type)

    def get(self, url, params=None, headers={}):
        url_parsed = urllib.parse.urlparse(url)
        path = url_parsed.path
        if params is not None:
            payload = urllib.parse.urlencode(params)
            path += f"?{payload}"
        self._send("GET", url_parsed.netloc, path, headers)

    def header(self, name):
        return self.response.getheader(name)



import re
import logging
from datetime import timedelta, timezone
from multiprocessing import Manager
from multiprocessing.pool import ThreadPool
from functools import partial

from yats.custom_datetime import CustomDateTime as datetime
from yats.twitter_request import TwitterRequest
from yats.profile import Profile
from yats.tweet_set import TweetSet
from yats.requests_holder import RequestsHolder

TWITTER_CREATION_DATE = datetime(2006, 3, 21)
# COUNT_QUERY = 1000
COUNT_QUERY = 20


class Connector:

    def __init__(self, fast_mode=True):
        logging.basicConfig(level=logging.DEBUG,
                            format=('%(message)s'))

    def profile(self, name, request=None):
        request = TwitterRequest() if request is None else request
        request.get_profile_request(name)
        res = request.body
        profile_res = Profile(res, verbose=True)
        return profile_res

    def _create_query(self,
                      q: str = None,
                      words: list = None,
                      sentence: str = None,
                      words_or: list = None,
                      words_not: list = None,
                      hashtag: str = None,
                      from_account: str = None,
                      to_account: str = None,
                      mention: str = None,
                      min_replies: int = None,
                      min_likes: int = None,
                      min_retweets: int = None,
                      since: datetime = None,
                      until: datetime = None,
                      filter_links: bool = None,
                      filter_replies: bool = None):
        if q is not None:
            query = f'{q} '
        else:
            query = ""
            if words is not None:
                query = f'{" ".join(words)} '
            if sentence is not None:
                query += f'"{sentence}" '
            if words_or is not None:
                query += f'({" OR ".join(words_or)}) '
            if words_not is not None:
                query += f'{" ".join(["-"+x for x in words_not])} '
            if hashtag is not None:
                query += f'({"#"+hashtag if hashtag[0] != "#" else hashtag}) '
            if from_account is not None:
                query += f'(from:{from_account}) '
            if to_account is not None:
                query += f'(to:{to_account}) '
            if mention is not None:
                query += f'({mention}) '
            if min_replies is not None:
                query += f'min_replies:{min_replies} '
            if min_likes is not None:
                query += f'min_faves:{min_likes} '
            if min_retweets is not None:
                query += f'min_retweets:{min_retweets} '
            if filter_links is not None and filter_links:
                query += "-filter:links "
            if filter_replies is not None and filter_replies:
                query += "-filter:replies "
        if until is not None:
            query += f'until:{until.strftime("%Y-%m-%d")} '
        if since is not None:
            query += f'since:{since.strftime("%Y-%m-%d")} '
        # check if query finishes by a space
        if query[-1] == " ":
            query = query[:-1]
        return query

    def _create_payload(self, query, count=COUNT_QUERY):
        return {
            "include_profile_interstitial_type": "1",
            "include_blocking": "1",
            "include_blocked_by": "1",
            "include_followed_by": "1",
            "include_want_retweets": "1",
            "include_mute_edge": "1",
            "include_can_dm": "1",
            "include_can_media_tag": "1",
            "skip_status": "1",
            "cards_platform": "Web-12",
            "include_cards": "1",
            "include_ext_alt_text": "true",
            "include_quote_count": "true",
            "include_reply_count": "1",
            "tweet_mode": "extended",
            "include_entities": "true",
            "include_user_entities": "true",
            "include_ext_media_color": "true",
            "include_ext_media_availability": "true",
            "send_error_codes": "true",
            "simple_quoted_tweet": "true",
            "q": query,
            "count": count,
            "query_source": "typed_query",
            "pc": "1",
            "spelling_corrections": "1",
            "ext": "mediaStats,highlightedLabel"
        }

    def _extract_since_until_from_q(self, q):
        until = None
        regex = r"until:(\d{4}-\d{2}-\d{2})"
        se = re.search(regex, q)
        if se:
            until = (datetime.strptime(se.group(1), "%Y-%M-%d")
                     .replace(tzinfo=timezone.utc))
            q = re.sub(regex, "", q)
        since = None
        regex = r"since:(\d{4}-\d{2}-\d{2})"
        se = re.search(regex, q)
        if se:
            since = (datetime.strptime(se.group(1), "%Y-%M-%d")
                     .replace(tzinfo=timezone.utc))
            q = re.sub(regex, "", q)
        return since, until, q

    def _tweet_worker(self, requests, lock, payload):
        with lock:
            if len(requests) > 0:
                request = requests.pop()
            else:
                request = TwitterRequest()
        tweets = TweetSet()
        data, cursor = request.get_tweets_request(payload)
        new_tweets = TweetSet(data)
        tweets.add(new_tweets)
        payload["cursor"] = cursor
        last_inserted = len(new_tweets)
        while last_inserted > 0:
            data, cursor = request.get_tweets_request(payload)
            new_tweets = TweetSet(data)
            tweets.add(new_tweets)
            payload["cursor"] = cursor
            last_inserted = len(new_tweets)
        with lock:
            requests.push(request)
        return tweets

    def _payload_generator(self,
                           q=None,
                           since=None,
                           until=None,
                           **args):
        def_since = TWITTER_CREATION_DATE
        def_until = (datetime.now(timezone.utc)
                     + timedelta(days=1))
        if q is not None:
            q_since, q_until, q = self._extract_since_until_from_q(q)
            since = q_since if q_since is not None else def_since
            until = q_until if q_until is not None else def_until
        else:
            since = def_since if since is None else since
            until = def_until if until is None else until
        beg_date = since
        end_date = beg_date + timedelta(days=1)
        while beg_date < until:
            query = self._create_query(q=q,
                                       since=beg_date,
                                       until=end_date,
                                       **args)
            payload = self._create_payload(query)
            yield payload
            beg_date = end_date
            end_date += timedelta(days=1)

    def get_all_tweets_request(self,
                               thread_nb=20,
                               **args):
        manager = Manager()
        requests = RequestsHolder()
        for _ in range(thread_nb):
            requests.push(TwitterRequest())
        lock = manager.Lock()
        tweets = TweetSet()
        with ThreadPool(thread_nb) as p:
            for new_tweets in p.imap_unordered(
                    partial(self._tweet_worker, requests, lock),
                    self._payload_generator(**args)):
                tweets.add(new_tweets)
                logging.debug(f"TOTAL {len(tweets)}, NEW {len(new_tweets)}")
        return tweets

    def get_tweets_user(self, username, since=None, **args):
        request = None
        if since is not None:
            beg_date = since
        else:
            request = TwitterRequest()
            profile = self.profile(username, request)
            beg_date = profile.creation
            logging.debug(f"Getting {profile.name}'s all tweets...")
        username = "@"+username if username[0] != "@" else ""
        tweets = self.get_all_tweets_request(from_account=username,
                                             since=beg_date,
                                             filter_replies=True,
                                             **args)
        return tweets




from yats.sorted_set import SortedSet
from yats.tweet import Tweet


class TweetSet(SortedSet):

    def __init__(self, data=[]):
        if not data or isinstance(data, list):
            super().__init__(data)
        else:
            super().__init__([])
            tweets_raw = data["tweets"]
            users_raw = data["users"]
            for id, content in tweets_raw.items():
                tweet = Tweet(content, users_raw)
                self.add(tweet)





class SortedElement:

    def __init__(self, sorted_keys):
        self.SORT_KEYS = sorted_keys

    def __eq__(self, other):
        return self.id == other.id

    def __lt__(self, other):
        for key in self.SORT_KEYS:
            if getattr(self, key) < getattr(other, key):
                return True
        return False

    def to_dict(self):
        return self.__dict__



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse

from yats.connector import Connector


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--username",
                        "-u",
                        type=str,
                        default="realdonaldtrump")
    parser.add_argument("--thread",
                        "-t",
                        type=int,
                        default=10)
    args = parser.parse_args()
    con = Connector(args.username)
    profile = con.profile(args.username)
    profile.describe()
    con.get_tweets_user(args.username, thread_nb=args.thread)


if __name__ == "__main__":
    main()



import sys
import bisect
import json
import csv
import re

from yats.serie import Serie
from yats.sorted_element import SortedElement
from yats.custom_datetime import CustomDateTime as datetime


class SortedSet:

    def __init__(self, init):
        self.list = list(init)

    def __len__(self):
        return len(self.list)

    def __iter__(self):
        return self.list.__iter__()

    def __next__(self):
        return self.list.__next__()

    def _get_list_attribute(self, att):
        return [getattr(x, att) for x in super().__getattribute__("list")]

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.list[item]
        elif isinstance(item, slice):
            return type(self)(self.list[item])
        elif isinstance(item, tuple):
            cols = []
            for it in item:
                cols.append(self._get_list_attribute(it))
            return Serie(zip(*cols))
        elif isinstance(item, Serie):
            init_list = [x[0] for x in zip(self.list, item) if x[1]]
            return type(self)(init_list)
        else:
            return Serie(self._get_list_attribute(item))

    def __getattribute__(self, name):
        try:
            attr = super().__getattribute__(name)
        except AttributeError:
            attr = Serie(self._get_list_attribute(name))
        return attr

    def add(self, data):
        if isinstance(data, SortedSet):
            beg = 0
            for element in data:
                idx = bisect.bisect(self.list, element, lo=beg)
                if 0 < idx < len(self.list) and element == self.list[idx]:
                    print("DUPLICATE", element)
                    print(element.id)
                    print(self.id)
                    sys.exit(0)
                else:
                    self.list.insert(idx, element)
                beg = idx
        else:
            idx = bisect.bisect(self.list, data)
            # bisect.insort(self.list, data)
            # return
            if 0 < idx < len(self.list) and data == self.list[idx]:
                pass
            else:
                self.list.insert(idx, data)

    def export(self, output):
        extension = re.search(r".+\.(\S+)$", output).group(1)
        if extension == "json":
            with open(output, 'w') as outfile:
                json.dump(self.list, outfile, cls=SortedListEncoder, indent=4)
        elif extension == "csv":
            with open(output, 'w', newline='') as csvfile:
                fieldnames = self.list[0].to_dict().keys()
                writer = csv.DictWriter(csvfile, fieldnames=fieldnames,
                                        quoting=csv.QUOTE_NONNUMERIC)

                writer.writeheader()
                for element in self.list:
                    writer.writerow(element.to_dict())


class SortedListEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, SortedElement):
            return o.to_dict()
        if isinstance(o, datetime):
            return o.isoformat()
        else:
            return json.JSONEncoder.default(self, o)





class RequestsHolder:

    def __init__(self):
        self.requests = []

    def __len__(self):
        return len(self.requests)

    def pop(self):
        return self.requests.pop()

    def push(self, request):
        self.requests.append(request)



from yats.custom_datetime import CustomDateTime as datetime

SUPPORTED_TYPES = (str, datetime, int, bool)


class Serie(list):

    def __eq__(self, other):
        if isinstance(other, SUPPORTED_TYPES):
            return Serie([x == other for x in self])
        else:
            return super().__eq__(other)

    def __ne__(self, other):
        if isinstance(other, SUPPORTED_TYPES):
            return Serie([x != other for x in self])
        else:
            return super().__ne__(other)

    def __ge__(self, other):
        if isinstance(other, SUPPORTED_TYPES):
            return Serie([x >= other for x in self])
        else:
            return super().__ge__(other)

    def __gt__(self, other):
        if isinstance(other, SUPPORTED_TYPES):
            return Serie([x > other for x in self])
        else:
            return super().__gt__(other)

    def __le__(self, other):
        if isinstance(other, SUPPORTED_TYPES):
            return Serie([x <= other for x in self])
        else:
            return super().__le__(other)

    def __lt__(self, other):
        if isinstance(other, SUPPORTED_TYPES):
            return Serie([x < other for x in self])
        else:
            return super().__lt__(other)

    def lc(self):
        return Serie([x.lower() for x in self if x is not None])

    def uc(self):
        return Serie([x.upper() for x in self if x is not None])



import logging

from yats.custom_datetime import CustomDateTime as datetime


class Profile:

    def __init__(self, response, verbose=False):
        user = response["data"]["user"]
        self.user = user
        self.name = user["legacy"]["name"]
        self.username = user["legacy"]["screen_name"]
        self.id = user["id"]
        self.restid = user["rest_id"]
        creation_str = user["legacy"]["created_at"]
        self.creation = datetime.strptime(creation_str,
                                          "%a %b %d %H:%M:%S %z %Y")
        self.description = user["legacy"]["description"]
        self.tweet_nb = user["legacy"]["statuses_count"]
        self.followers_nb = user["legacy"]["followers_count"]
        self.following_nb = user["legacy"]["friends_count"]
        self.favourites_nb = user["legacy"]["favourites_count"]

    def __repr__(self):
        return f"<pyter.Profile:{self.username}>"

    def describe(self):
        logging.debug(f"{self.name}'s Profile:")
        logging.debug(f"{self.username} {self.id}")
        logging.debug(f"{self.description}")
        logging.debug("Twitter Account:")
        logging.debug(f"Creation date: {self.creation}")
        logging.debug(f"Number of Tweets: {self.tweet_nb}")
        logging.debug(f"Number of followers: {self.followers_nb}")
        logging.debug(f"Number of following: {self.following_nb}")
        logging.debug(f"Number of favourites: {self.favourites_nb}")



import re
import sys
import logging

from yats.request import Request

TWITTER_URL = "https://mobile.twitter.com/"
GRAPHQL_URL = "https://api.twitter.com/graphql/"
USER_INFO_QL_PATH = "UserByScreenName"
USER_INFO_QL_ID = "-xfUfZsnR_zqjFd-IfrN5A"
USER_AGENT = {
    "User-Agent": ("Mozilla/5.0 (X11; Linux x86_64) "
                   "AppleWebKit/537.36 (KHTML, like Gecko) "
                   "Chrome/83.0.4103.97 Safari/537.36"),
}
TOKEN_BEARER = ("AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs"
                "%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA")


class TwitterRequest(Request):

    def __init__(self, fast_mode=True):
        if fast_mode:
            self.graphql_ext = USER_INFO_QL_ID
            self.token_bearer = TOKEN_BEARER
        self.token_guest = None
        super().__init__()

    def _is_unset(self, attributes):
        if isinstance(attributes, list):
            for attribute in attributes:
                if not hasattr(self, attribute) or attribute is None:
                    return True
            return False
        else:
            return not hasattr(self, attribute) or attribute is None

    def _get_initial_request(self):
        logging.debug("Getting the token_guest and main_js...")
        logging.debug(f"  get initial request {TWITTER_URL}")
        self.get(TWITTER_URL, headers=USER_AGENT)
        self.token_guest = re.search(r"gt=(\w+)",
                                     self.body).group(1)
        self.main_js = re.search(r"https://abs.twimg.com/responsive-web/web/"
                                 r"(main.(?:.*?)\.js)",
                                 self.body).group(0)

    def _get_main_js_page(self):
        logging.debug("Getting the token_bearer and graphql_ext...")
        self.get(self.main_js)
        self.token_bearer = re.search(r's=\"([a-zA-Z0-9+%]{10,}?)\"',
                                      self.body).group(1)
        regex = r"queryId:\"(.*?)\",operationName:\"(.*?)\""
        queries = re.findall(regex, self.body)
        pair = next(x for x in queries if x[1] == USER_INFO_QL_PATH)
        self.graphql_ext = pair[0]

    def _get_connection_infos(self):
        if self._is_unset(["token_guest", "main_js"]):
            self._get_initial_request()
        if self._is_unset(["graphql_ext", "token_bearer"]):
            self._get_main_js_page()

    def _get_cursor(self, timeline):
        instructions = timeline["instructions"]
        for instruction in instructions:
            if "addEntries" in instruction.keys():
                for entry in instruction["addEntries"]["entries"]:
                    if entry["entryId"] == "sq-cursor-bottom":
                        content = entry["content"]
                        cursor = content["operation"]["cursor"]["value"]
                        return cursor
        for instruction in instructions:
            if "replaceEntry" in instruction.keys():
                replace = instruction["replaceEntry"]
                if replace["entryIdToReplace"] == "sq-cursor-bottom":
                    content = replace["entry"]["content"]
                    cursor = content["operation"]["cursor"]["value"]
                    return cursor
        else:
            cursor = None
            logging.error("No new cursor")
            sys.exit(0)
        return cursor

    def get_profile_request(self, username):
        self._get_connection_infos()
        user_url = GRAPHQL_URL + self.graphql_ext + "/" + USER_INFO_QL_PATH
        logging.debug(f"Getting {username} infos...")
        payload = {
            "variables": (f'{{"screen_name":"{username}",'
                          f'"withHighlightedLabel":true}}')
        }
        headers = {
            "authorization": f"Bearer {self.token_bearer}"
        }
        self.get(user_url, headers=headers, params=payload)

    def get_tweets_request(self, payload):
        self._get_connection_infos()
        headers = {
            "authorization": f"Bearer {self.token_bearer}",
            "x-csrf-token": "dbfeef183c6a3f1f4f1609aa22f3b378",
            "x-guest-token": self.token_guest
        }
        self.get("https://api.twitter.com/2/search/adaptive.json",
                 params=payload,
                 headers=headers)
        data = self.body["globalObjects"]
        cursor = self._get_cursor(self.body["timeline"])
        rate_limit = int(self.header("x-rate-limit-remaining"))
        if rate_limit < 10:
            logging.warning(f"WARNING, rate-limit = {rate_limit}")
            logging.warning("Refreshing now")
            self._get_initial_request()
        return data, cursor



import sys

from yats.custom_datetime import CustomDateTime as datetime
from yats.sorted_element import SortedElement
SORT_KEYS = ["creation"]


class Tweet(SortedElement):

    def __init__(self, data, users, sort_keys=SORT_KEYS):
        super().__init__(sort_keys)
        creation_str = data["created_at"]
        self.creation = datetime.strptime(creation_str,
                                          "%a %b %d %H:%M:%S %z %Y")
        self.id = data["id"]
        self.id_str = data["id_str"]
        self.text = data["full_text"]
        self.truncated = data["truncated"]
        self.hashtags = data["entities"]["hashtags"]
        self.symbols = data["entities"]["symbols"]
        self.mentions = data["entities"]["user_mentions"]
        self.urls = data["entities"]["urls"]
        self.source = data["source"]
        self.reply_to = data["in_reply_to_user_id_str"]
        self.user_id = data["user_id"]
        self.user_id_str = data["user_id_str"]
        self.username = users[self.user_id_str]["screen_name"]
        self.localisation = data["geo"]
        self.retweet_nb = data["retweet_count"]
        self.favorite_nb = data["favorite_count"]
        self.reply_nb = data["reply_count"]
        self.lang = data["lang"]

    def __repr__(self):
        return (f"<pyter.Tweet"
                f":{self.creation.strftime('%y-%m-%d %H:%M:%S')}"
                f":{self.id}"
                f":{self.username}>")

    def to_dict(self):
        d = dict(self.__dict__)
        del d["SORT_KEYS"]
        return d
